#include "proplist.h"

// Child window IDs
//
#define PB_EDIT_ID      1203
#define PB_DDBUTTON_ID  1204
#define PB_ODBUTTON_ID  1205

// Margin for value area of property box
//
#define kValueMargin    2

// Default width for buttons
//
#define kDefButtonWidth 18


// TPropNotify implementation
//
TPropNotify::TPropNotify(HWND ctl, uint id, uint code, TPropertyItem* item,
                         TRect& rect)
:
 TNotify(ctl, id, code), pItem(item), ownerDropRect(rect)
{
}

// TPropListBox implementation
//
DEFINE_RESPONSE_TABLE1(TPropListBox, TListBox)
  EV_NOTIFY_AT_CHILD(LBN_SELCHANGE, EvSelChange),
  EV_WM_DRAWITEM,
  EV_WM_SETFOCUS,
  EV_WM_SIZE,
  EV_WM_VSCROLL,
END_RESPONSE_TABLE;

TPropListBox::TPropListBox(TWindow* parent, int resourceId, TModule* module)
:
 TListBox(parent, resourceId, module),
 listBox(0),
 textColor           (TColor::SysWindowText),
 selTextColor        (TColor::SysHighlightText),
 readOnlyTextColor   (TColor::SysGrayText),
 gridColor           (TColor::SysGrayText),
 propertyBackColor   (TColor::Sys3dFace),
 selPropertyBackColor(TColor::SysHighlight),
 valueBackColor      (224, 224, 224)
{
  // Build the floating edit box
  //
  edit = new TSubRichEdit(this, PB_EDIT_ID, "", 0, 0, 0, 0);

  // Build the drop down button
  //
  button = new TDropDownButton(this, PB_DDBUTTON_ID);

  // Build the owner drop button
  //
  odButton = new TOwnerDropButton(this, PB_ODBUTTON_ID);

  // Grid pen
  //
  gridPen = new TPen(gridColor);

  valueRect.SetNull();
}

TPropListBox::~TPropListBox()
{
  delete edit;
  delete button;
  delete odButton;

  delete gridPen;

  strToItemMap.Flush();
}

TPropertyItem*
TPropListBox::GetProperty(int index)
{
  if (index < 0 || index >= GetCount())
    return 0;

  char temp[128];
  GetString(temp, index);

  return GetProperty(temp);
}

TPropertyItem*
TPropListBox::GetProperty(const char* property)
{
  TPropertyItem* item;

  if (strToItemMap.GetItemsInContainer()) {
    TStrToItemAssoc* ptr = strToItemMap.Find(TStrToItemAssoc(property, 0));
    if (ptr)
      item = const_cast<TPropertyItem*>(ptr->Value());
    else
      item = 0;
  }
  else
    item = 0;

  return item;
}

int
TPropListBox::AddProperty(TPropertyItem* item)
{
  if (!item)
    return -1;

  const char* property = item->GetProperty();

  // Each property name can only appear once
  //
  TPropertyItem* lookup = GetProperty(property);
  if (lookup) {
    delete item;

    return LB_ERR;
  }

  // Check for invalid flag combination
  //
  int flags = item->GetFlags();

  if (flags & PROP_DROPLIST)
    if (flags & PROP_OWNERDROP || flags & PROP_READONLY)
      return -1;

  if (flags & PROP_OWNERDROP)
    if (flags & PROP_DROPLIST || flags & PROP_READONLY)
      return -1;

  if (flags & PROP_READONLY)
    if (flags & PROP_DROPLIST || flags & PROP_OWNERDROP)
      return -1;

  strToItemMap.Add(TStrToItemAssoc(property, item));

  return AddString(property);
}

int
TPropListBox::DelProperty(const char* property)
{
  int index = FindString(property, -1);
  if (index >= 0) {
    HideChildren();

    index = DeleteString(index);

    TPropertyItem* item = GetProperty(property);
    if (item) {
      TStrToItemAssoc* ptr = strToItemMap.Find(TStrToItemAssoc(property, 0));
      strToItemMap.Detach(*ptr);
    }
  }

  SetSelIndex(index);

  Invalidate();

  return index;
}

bool
TPropListBox::SetValue(TPropertyItem* item, const char* value)
{
  string oldValue = item->GetValue();

  item->SetValue(value);

  if (item->GetFlags() & PROP_GETCHANGE)
    // Notify of the property value change
    //
    if (SendPropChange(item)) {
      item->SetValue(oldValue.c_str());

      return false;
    }

  int index = FindString(item->GetProperty(), -1);
  if (index == GetSelIndex())
    edit->SetWindowText(value);

  TRect rect;
  GetItemRect(index, rect);

  InvalidateRect(rect);

  return true;
}

void
TPropListBox::SetPropertyWidth(int width)
{
  int difference = width - propertyWidth;

  if (difference) {
    propertyWidth = width;

    valueRect.left += difference;

    Invalidate(true);

    EvSelChange();
  }
}

void
TPropListBox::EvSelChange()
{
  HideChildren();

  // Get the newly selected property
  //
  TPropertyItem* item = GetProperty(GetSelIndex());
  if (!item)
    return;

  int flags = item->GetFlags();

  // Update and reposition the floating child windows
  //
  edit->SetWindowText(item->GetValue());

  if (flags & PROP_READONLY)
    return;

  TRect editRect = valueRect;
  editRect.left -= kValueMargin;

  // Handle the buttons
  //
  if (flags & PROP_DROPLIST || flags & PROP_OWNERDROP) {
    TRect buttonRect;
    GetItemRect(GetSelIndex(), buttonRect);

    buttonRect.right--;
    buttonRect.bottom--;

    editRect.right -= kDefButtonWidth;

    buttonRect.left = editRect.right;

    TButton* ptr;
    if (flags & PROP_OWNERDROP)
      ptr = odButton;
    else
      ptr = button;

    ptr->MoveWindow(buttonRect);
    ptr->ShowWindow(SW_SHOW);

    listBox = item->GetListBox();

    if (listBox)
      listBox->SetWindowFont(Parent->GetWindowFont(), true);
  }

  // If the PROP_EDIT flag is set, allow the user to edit
  //
  edit->AllowEdit(flags & PROP_EDIT ? true : false);

  edit->MoveWindow(editRect);
  edit->ShowWindow(SW_SHOW);
}

TResult
TPropListBox::EvCommand(uint id, THandle hWndCtl, uint notifyCode)
{
  char temp[1024];

  TPropertyItem* item = GetProperty(GetSelIndex());
  if (!item)
    return TListBox::EvCommand(id, hWndCtl, notifyCode);

  switch (notifyCode) {
    case BN_CLICKED:
      // A child button was clicked, toggle the drop down list box
      //
      if (id == PB_DDBUTTON_ID || id == PB_ODBUTTON_ID) {
        SendPressButton(item);

        ToggleList(item);

        return 1;
      }
      break;

    case LBN_DBLCLK:
      // The child list box was double clicked, set the value
      //
      if (id == PB_LISTBOX_ID) {
        int index = listBox->GetSelIndex();

        listBox->GetString(temp, index);
        listBox->Show(SW_HIDE);

        SetValue(item, temp);

        SetFocus();

        return 1;
      }
      break;

    case SN_KILLFOCUS:
    case SN_RETURN: {
      // The enter key was pressed, set the value
      //
      string oldValue(item->GetValue());

      int index    = -1;
      int oldIndex = item->GetIndex();

      if (id == PB_EDIT_ID)
        edit->GetWindowText(temp, sizeof(temp));
      else {
        index = listBox->GetSelIndex();
        if (index == -1)
          return 0;

        listBox->GetString(temp, index);
        listBox->ShowWindow(SW_HIDE);

        edit->SetWindowText(temp);
      }

      item->SetValue(temp);
      item->SetIndex(index);

      if (item->GetFlags() & PROP_GETCHANGE)
        // Notify of the property value change
        //
        if (SendPropChange(item)) {
          item->SetValue(const_cast<char *>(oldValue.c_str()));
          item->SetIndex(oldIndex);

          edit->SetWindowText(oldValue.c_str());

          return 0;
        }

      SetFocus();

      return 1;
    }

    case SN_ESCAPE:
      // The escape key was pressed, undo the changes
      //
      if (id == PB_EDIT_ID)
        edit->SetWindowText(item->GetValue());
      else {
        listBox->SetSelString(item->GetValue(), -1);
        listBox->ShowWindow(SW_HIDE);
      }

      SetFocus();

      return 1;

    // Scroll up or down the list of values
    //
    case SN_DOWN:
    case SN_UP:
      if (listBox && listBox->GetCount() > 0) {
        edit->GetWindowText(temp, sizeof(temp));

        int index     = listBox->SetSelString(temp, -1);
        int lastIndex = listBox->GetCount() - 1;

        if (notifyCode == SN_DOWN) {
          if (index < lastIndex)
            index++;
        }
        else if (index > 0)
          index--;

        listBox->GetString(temp, index);

        if (GetSelIndex() >= 0) {
          SetValue(item, temp);

          edit->SetWindowText(temp);
          edit->SetSelection(0, 0);
        }
      }

      return 1;
  }

  return TListBox::EvCommand(id, hWndCtl, notifyCode);
}

bool
TPropListBox::PreProcessMsg(MSG& msg)
{
  WPARAM wParam = msg.wParam;

  switch (msg.message) {
    case WM_KEYDOWN:
      // Handle normal keys for control navigation
      //
      if (msg.hwnd == GetHandle()) {
        if (wParam == VK_DOWN || wParam == VK_UP) {
          int oldIndex = GetSelIndex();

          if (oldIndex != -1) {
            int newIndex = oldIndex;

            if (wParam == VK_DOWN) {
              if (newIndex < (GetCount() - 1))
                newIndex++;
            }
            else if (newIndex > 0)
              newIndex--;

            if (newIndex != oldIndex) {
              SetSelIndex(newIndex);

              EvSelChange();
            }
          }
        }

        return true;
      }

      break;

    case WM_SYSKEYDOWN: {
      TPropertyItem* item = GetProperty(GetSelIndex());

      if (item) {
        // Handle the ALT key for control navigation
        //
        if (wParam == VK_MENU) {
          if (!(item->GetFlags() & PROP_READONLY)) {
            THandle focus = GetFocus();

            if (focus != edit->HWindow &&
                (!listBox || focus != listBox->GetHandle()))
              edit->SetFocus();
          }
        }
        else if ((wParam == VK_DOWN || wParam == VK_UP) && listBox) {
          ListState state;

          if (wParam == VK_DOWN) {
            listBox->SetFocus();

            state = show;
          }
          else
            state = hide;

          ToggleList(item, state);
        }

        return true;
      }

      break;
    }
  }

  return TListBox::PreProcessMsg(msg);
}

void
TPropListBox::SetupWindow()
{
  TListBox::SetupWindow();

  SetBkgndColor(TColor(192, 192, 192));

  TFont font(Parent->GetWindowFont());
  SetWindowFont(font, true);

  // Create the edit box
  //
  edit->Create();

  // And set the font for it
  //
  LOGFONT logFont;
  font.GetObject(logFont);

  logFont.lfWeight = FW_BOLD;

  TFont editFont(&logFont);
  edit->SetWindowFont(editFont, true);

  // Create the drop down button
  //
  button->Create();

  // Create the owner drop button
  //
  odButton->Create();
}

void
TPropListBox::EvDrawItem(uint /* ctrlId */, DRAWITEMSTRUCT& drawInfo)
{
  TDC dc(drawInfo.hDC);

  RECT& rcItem = drawInfo.rcItem;

  TPropertyItem* item = GetProperty(drawInfo.itemID);

  // Is the item read only?
  //
  bool isReadOnly = item->GetFlags() & PROP_READONLY;

  // Is the item selected?
  //
  bool isSelected = drawInfo.itemState & ODS_SELECTED;

  // Calculate the dimensions of the rectangles surrounding
  // the property and value parts
  //
  TRect propertyRect(rcItem);
  propertyRect.right = propertyRect.left + propertyWidth;

  valueRect = rcItem;
  valueRect.left = propertyRect.right;

  // Draw the property and value rectangles
  //
  dc.FillRect(propertyRect,
              TBrush(isSelected ? selPropertyBackColor : propertyBackColor));

  if (!isSelected || isReadOnly)
    dc.FillRect(valueRect, TBrush(valueBackColor));

  // Adjust the rectangles for text fitness
  //
  propertyRect.Inflate(-kValueMargin, 0);
  valueRect.left += kValueMargin;

  // Specify the format for text representation
  //
  uint16 format = DT_LEFT | DT_SINGLELINE | DT_NOPREFIX;

  dc.SetBkMode(TRANSPARENT);

  // Draw the text for property
  //
  dc.SetTextColor(isSelected ? selTextColor : textColor);

  dc.DrawText(item->GetProperty(), -1, propertyRect, format);

  // Draw the text for value
  //
  dc.SetTextColor(isReadOnly ? readOnlyTextColor : textColor);

  if (isSelected && !isReadOnly)
    valueRect.bottom--;
  else
    dc.DrawText(item->GetValue(), -1, valueRect, format);

  // Draw the grid lines
  //
  dc.SelectObject(*gridPen);

  int temp;

  temp = rcItem.bottom - 1;
  dc.MoveTo(rcItem.left  - 1, temp);
  dc.LineTo(rcItem.right + 1, temp);

  temp = propertyRect.right + 1;
  dc.MoveTo(temp, propertyRect.bottom - 1);
  dc.LineTo(temp, propertyRect.top    - 1);

  dc.RestoreObjects();
}

void
TPropListBox::EvSetFocus(THandle hWndLostFocus)
{
  TListBox::EvSetFocus(hWndLostFocus);

  edit->SetSelection(0, 0);

  button->BringWindowToTop();

  if (listBox)
    listBox->ShowWindow(SW_HIDE);
}

void
TPropListBox::EvSize(uint sizeType , TSize& size)
{
  TListBox::EvSize(sizeType, size);

  EvSelChange();
}

void
TPropListBox::EvVScroll(uint scrollCode, uint thumbPos, HWND hWndCtl)
{
  if (GetSelIndex() != -1) {
    // Hide floating controls and unselect
    // the item at current index position
    //
    HideChildren();

    SetSelIndex(-1);
  }

  TListBox::EvVScroll(scrollCode, thumbPos, hWndCtl);
}

bool
TPropListBox::SendNotification(uint code, TPropertyItem* item, TRect& rect)
{
  if (!Parent)
    return false;

  // Build the notification and send the message
  //
  TPropNotify nd(GetHandle(), GetDlgCtrlID(), code, item, rect);

  return Parent->SendNotification(GetSelIndex(), nd);
}

void
TPropListBox::ToggleList(TPropertyItem* item, ListState state)
{
  TRect itemRect;
  GetItemRect(GetSelIndex(), itemRect);

  if (item->GetFlags() & PROP_DROPLIST) {
    // If the list is visible, hide it and commit the value
    //
    if (listBox->IsWindowVisible()) {
      if (state == neutral || state == hide) {
        int index = listBox->GetSelIndex();

        if (index >= 0) {
          char temp[1024];
          listBox->GetString(temp, index);

          item->SetValue(temp);

          edit->SetWindowText(temp);
        }

        listBox->ShowWindow(SW_HIDE);

        if (edit->HWindow && edit->IsWindowVisible())
          edit->SetFocus();
      }
    }
    else if (state == neutral || state == show) {
      TRect clientRect;
      GetClientRect(clientRect);

      TRect rect = itemRect;

      rect.left += propertyWidth;
      rect.top   = rect.bottom;

      edit->SetSelection(0, 0);

      int listHeight = listBox->GetItemHeight(0) * (listBox->GetCount() + 1);
      rect.bottom += listHeight;

      if (rect.bottom > clientRect.bottom) {
        int top = itemRect.top;

        int upperHeight = top - clientRect.top;

        if (listHeight <= upperHeight) {
          rect.bottom = top;
          rect.top    = top - listHeight;
        }
        else
          rect.bottom = clientRect.bottom;
      }

      if (listBox->SetSelString(item->GetValue(), -1) == -1)
        listBox->SetSelIndex(-1);

      listBox->MoveWindow(rect);
      listBox->ShowWindow(SW_SHOW);
      listBox->SetFocus();
    }
  }
  else if (item->GetFlags() & PROP_OWNERDROP)
    // The application must handle the drop down list
    //
    SendOwnerDrop(item, itemRect);
}

void
TPropListBox::HideChildren()
{
  edit    ->ShowWindow(SW_HIDE);
  button  ->ShowWindow(SW_HIDE);
  odButton->ShowWindow(SW_HIDE);

  if (listBox)
    listBox->ShowWindow(SW_HIDE);
}

